<?php

namespace Drupal\pwa_firebase;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Path\CurrentPathStack;
use GuzzleHttp\ClientInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Theme\ThemeManagerInterface;

/**
 * Service to send notification.
 */
class Firebase implements FirebaseInterface {

  /**
   * Current path stack.
   *
   * @var \Drupal\Core\Path\CurrentPathStack
   */
  protected CurrentPathStack $currentPathStack;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected ConfigFactoryInterface $configFactory;

  /**
   * HTTP client.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected ClientInterface $client;

  /**
   * Database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  private Connection $connection;

  /**
   * The Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected MessengerInterface $messenger;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected ModuleHandlerInterface $moduleHandler;

  /**
   * The entity services.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The logger factory service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected LoggerChannelFactoryInterface $loggerFactory;

  /**
   * Firebase configuration.
   *
   * @var mixed
   */
  protected $firebaseConfig;

  /**
   * Theme manager service.
   *
   * @var \Drupal\Core\Theme\ThemeManagerInterface
   */
  protected ThemeManagerInterface $themeManager;

  const TITLE_LENGTH = 255;
  // Assume that these are now config values.
  const MESSAGE_LENGTH = 1024;

  /**
   * Constructs a FirebaseServiceBase object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The factory for configuration objects.
   * @param \GuzzleHttp\ClientInterface $client
   *   An HTTP client.
   * @param \Drupal\Core\Database\Connection $connection
   *   Database connection.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity service.
   * @param \Drupal\Core\Path\CurrentPathStack $currentPathStack
   *   Current patch Stack.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   Logger Factory.
   * @param \Drupal\Core\Theme\ThemeManagerInterface $theme_manager
   *   Serializer data.
   *
   * @throws \Exception
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    ClientInterface $client,
    Connection $connection,
    MessengerInterface $messenger,
    ModuleHandlerInterface $module_handler,
    EntityTypeManagerInterface $entity_type_manager,
    CurrentPathStack $currentPathStack,
    LoggerChannelFactoryInterface $logger_factory,
    ThemeManagerInterface $theme_manager,
  ) {
    $this->configFactory = $config_factory;
    $this->client = $client;
    $this->connection = $connection;
    $this->messenger = $messenger;
    $this->moduleHandler = $module_handler;
    $this->entityTypeManager = $entity_type_manager;
    $this->currentPathStack = $currentPathStack;
    $this->loggerFactory = $logger_factory;
    $this->themeManager = $theme_manager;
    $this->firebaseConfig = $this->configFactory->get('pwa_firebase.settings');
  }

  /**
   * Function to send a notification to all the users.
   *
   * @param string $title
   *   Message title.
   * @param string $message
   *   Message body.
   * @param mixed $url
   *   Url to article.
   * @param mixed $option
   *   Option for device type.
   */
  public function sendMessageToAllUsers($title = '', $message = '', $url = NULL, $option = []) {
    $rows = $this->getFirebaseTokens();
    return $this->prepareSend($rows, $title, $message, $url, $option);
  }

  /**
   * Function to send a notification to 1 user.
   */
  public function sendMessageToUser($uid, $title, $message, $url = NULL, $option = []) {
    $rows = $this->getFirebaseTokens(['uid' => $uid]);

    if (empty($rows)) {
      return FALSE;
    }

    $option['uid'] = $uid;
    $option['user_name'] = $this->getUserName($uid);

    return $this->prepareSend($rows, $title, $message, $url, $option);
  }

  /**
   * {@inheritDoc}
   */
  private function prepareSend(array $rows, string $title, string $message, $url = NULL, array $option = []): bool {
    $tokens = [];
    foreach ($rows as $row) {
      $token = $row->token;
      $device_type = 'pc';
      if (!empty($row->device)) {
        $device_type = $row->device;
      }
      if (!isset($tokens[$device_type])) {
        $tokens[$device_type] = [];
      }
      $tokens[$device_type][] = $token;
    }
    if (!empty($tokens)) {
      foreach ($tokens as $device => $token) {
        try {
          $option['device_type'] = $device;
          $this->sendNotification($token, $title, $message, $url, $option);
        }
        catch (\Exception $e) {
          $this->handleException($e);
        }
      }
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Function send message to firebase.
   */
  public function sendNotification($token, $title, $message, $url = NULL, $option = []) {
    global $base_url;
    $title = $this->trimText($title, self::TITLE_LENGTH);
    $message = $this->trimText($message, self::MESSAGE_LENGTH);
    $url = empty($url) ? $this->currentPathStack->getPath() : $url;
    $icon = $this->getNotificationIcon($option, $base_url, $url);
    $this->moduleHandler->alter('pwa_firebase_send', $option);
    $notification = $this->prepareNotification($title, $message, $icon, $url, $option);
    $data = $this->prepareData($notification, $token, $option);

    try {
      $headers = [
        'Content-Type' => 'application/json',
        'Accept' => 'application/json',
        'Authorization' => 'key=' . $this->firebaseConfig->get('firebase_server_key'),
      ];
      return $this->postNotificationData($this->firebaseConfig->get('firebase_endpoint'), $headers, $data, $option, $token);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return [];
    }
  }

  /**
   * {@inheritdoc}
   */
  private function trimText($text, $limit) {
    if (mb_strlen($text) > $limit) {
      $text = mb_substr($text, 0, $limit - 1) . '…';
    }
    return $text;
  }

  /**
   * {@inheritdoc}
   */
  private function getNotificationIcon($option, $base_url, $base_path) {
    $logo = $this->themeManager->getActiveTheme()->getLogo();
    $icon = !empty($option['image']) ? $base_url . $base_path . $option['image'] : $logo;
    if (filter_var($icon, FILTER_VALIDATE_URL) == FALSE) {
      $icon = $base_url . $icon;
    }
    return $icon;
  }

  /**
   * {@inheritdoc}
   */
  private function prepareNotification($title, $message, $icon, $url, &$option) {
    $notification = [
      'body' => $message,
      'title' => $title,
      'icon' => $icon,
      'click_action' => $url,
    ];
    foreach ($notification as $key => $value) {
      if (!empty($option[$key])) {
        $notification[$key] = $option[$key];
        unset($option[$key]);
      }
    }
    return $notification;
  }

  /**
   * {@inheritdoc}
   */
  private function prepareData($notification, $token, $option) {
    $data = [
      "content_available" => empty($option['content_available']) ? TRUE : FALSE,
      "priority" => !empty($option['priority']) ? $option['priority'] : 'high',
      "notification" => $notification,
      "data" => $option,
    ];
    if (is_array($token)) {
      if (count($token) == 1) {
        $data["to"] = current($token);
      }
      else {
        $data["registration_ids"] = $token;
      }
    }
    else {
      $data["to"] = $token;
    }
    return $data;
  }

  /**
   * {@inheritdoc}
   */
  private function postNotificationData($endpoint, $headers, $data, $option, $token) {
    $result = FALSE;
    try {
      $response = $this->client->post($endpoint, [
        'headers' => $headers,
        'body' => json_encode($data),
      ]);
      $result = json_decode($response->getBody(), true);
    }
    catch (\Exception $e) {
      $this->loggerFactory->get('firebase')->error($e->getMessage());
    }

    if (!empty($result["failure"])) {
      foreach ($result["results"] as $index => $value) {
        if (!empty($value['error'])) {
          $this->loggerFactory->get('firebase')->error($value['error'] . ': ' . ($option['user_name'] ?? '') . ' - ' . ($option['device'] ?? ''));
          $this->connection->delete('pwa_firebase')->condition('token', $token[$index])->execute();
        }
      }
    }
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  protected function handleException($e): void {
    $this->loggerFactory->get('firebase')->error($e->getMessage());
    $this->messenger->addError($this->t('There was an error sending the notification.'));
  }

  /**
   * {@inheritdoc}
   */
  protected function getFirebaseTokens(array $conditions = []): array {
    $tokens = [];
    try {
      $query = $this->connection->select('pwa_firebase', 'f')
        ->fields('f', ['token', 'device']);
      foreach ($conditions as $field => $value) {
        $query->condition($field, $value);
      }
      $tokens = $query->execute()->fetchAll();
    }
    catch (\Exception $e) {
      $this->handleException($e);
    }

    return $tokens;
  }

  /**
   * {@inheritdoc}
   */
  protected function getUserName(int $uid): ?string {
    $user = $this->entityTypeManager->getStorage('user')->load($uid);
    return $user ? $user->get('name')->value : NULL;
  }

}
